@@ -6,33 +6,50 @@
 
 #include "rimeengine.h"
 #include "notifications_public.h"
+#include "rimeaction.h"
 #include "rimestate.h"
 #include <cstdint>
 #include <cstring>
 #include <ctime>
 #include <dirent.h>
+#include <exception>
+#include <fcitx-config/iniparser.h>
+#include <fcitx-config/rawconfig.h>
 #include <fcitx-utils/event.h>
 #include <fcitx-utils/fs.h>
 #include <fcitx-utils/i18n.h>
 #include <fcitx-utils/log.h>
+#include <fcitx-utils/macros.h>
 #include <fcitx-utils/misc.h>
 #include <fcitx-utils/standardpath.h>
 #include <fcitx-utils/stringutils.h>
+#include <fcitx/action.h>
+#include <fcitx/addoninstance.h>
 #include <fcitx/candidatelist.h>
+#include <fcitx/event.h>
 #include <fcitx/inputcontext.h>
 #include <fcitx/inputcontextmanager.h>
+#include <fcitx/inputmethodentry.h>
 #include <fcitx/inputpanel.h>
+#include <fcitx/instance.h>
+#include <fcitx/menu.h>
 #include <fcitx/statusarea.h>
 #include <fcitx/userinterface.h>
 #include <fcitx/userinterfacemanager.h>
-#include <optional>
+#include <list>
+#include <memory>
 #include <rime_api.h>
 #include <stdexcept>
 #include <string>
+#include <string_view>
+#include <thread>
+#include <unordered_map>
+#include <utility>
+#include <vector>
 
-FCITX_DEFINE_LOG_CATEGORY(rime, "rime");
+FCITX_DEFINE_LOG_CATEGORY(rime_log, "rime");
 
-namespace fcitx {
+namespace fcitx::rime {
 
 namespace {
 
@@ -42,19 +59,21 @@ parseAppOptions(rime_api_t *api, RimeConfig *config) {
         appOptions;
     RimeConfigIterator appIter;
     RimeConfigIterator optionIter;
-    api->config_begin_map(&appIter, config, "app_options");
-    while (api->config_next(&appIter)) {
-        auto &options = appOptions[appIter.key];
-        api->config_begin_map(&optionIter, config, appIter.path);
-        while (api->config_next(&optionIter)) {
-            Bool value = False;
-            if (api->config_get_bool(config, optionIter.path, &value)) {
-                options[optionIter.key] = !!value;
+    if (api->config_begin_map(&appIter, config, "app_options")) {
+        while (api->config_next(&appIter)) {
+            auto &options = appOptions[appIter.key];
+            if (api->config_begin_map(&optionIter, config, appIter.path)) {
+                while (api->config_next(&optionIter)) {
+                    Bool value = False;
+                    if (api->config_get_bool(config, optionIter.path, &value)) {
+                        options[optionIter.key] = !!value;
+                    }
+                }
+                api->config_end(&optionIter);
             }
         }
-        api->config_end(&optionIter);
+        api->config_end(&appIter);
     }
-    api->config_end(&appIter);
     return appOptions;
 }
 
@@ -66,6 +85,7 @@ std::vector<std::string> getListItemPath(rime_api_t *api, RimeConfig *config,
         while (api->config_next(&iter)) {
             paths.push_back(iter.path);
         }
+        api->config_end(&iter);
     }
     return paths;
 }
@@ -75,7 +95,7 @@ std::vector<std::string> getListItemString(rime_api_t *api, RimeConfig *config,
     std::vector<std::string> values;
     auto paths = getListItemPath(api, config, path);
     for (const auto &path : paths) {
-        auto value = api->config_get_cstring(config, path.c_str());
+        const auto *value = api->config_get_cstring(config, path.c_str());
         if (!value) {
             return {};
         }
@@ -100,7 +120,7 @@ public:
 
     std::string shortText(InputContext *ic) const override {
         std::string result;
-        auto state = engine_->state(ic);
+        auto *state = engine_->state(ic);
         if (state) {
             state->getStatus([&result](const RimeStatus &status) {
                 result = status.schema_id ? status.schema_id : "";
@@ -122,7 +142,7 @@ public:
 
     std::string longText(InputContext *ic) const override {
         std::string result;
-        auto state = engine_->state(ic);
+        auto *state = engine_->state(ic);
         if (state) {
             state->getStatus([&result](const RimeStatus &status) {
                 result = status.schema_name ? status.schema_name : "";
@@ -133,7 +153,7 @@ public:
 
     std::string icon(InputContext *ic) const override {
         bool isDisabled = false;
-        auto state = engine_->state(ic);
+        auto *state = engine_->state(ic);
         if (state) {
             state->getStatus([&isDisabled](const RimeStatus &status) {
                 isDisabled = status.is_disabled;
@@ -149,149 +169,13 @@ private:
     RimeEngine *engine_;
 };
 
-class ToggleAction : public RimeOptionAction {
-public:
-    ToggleAction(RimeEngine *engine, std::string schema, std::string option,
-                 std::string disabledText, std::string enabledText)
-        : engine_(engine), option_(option), disabledText_(disabledText),
-          enabledText_(enabledText) {
-        engine_->instance()->userInterfaceManager().registerAction(
-            stringutils::concat("fcitx-rime-", schema, "-", option), this);
-    }
-
-    void activate(InputContext *ic) override {
-        auto state = engine_->state(ic);
-        auto api = engine_->api();
-        if (!state) {
-            return;
-        }
-        // Do not send notification since user is explicitly select it.
-        engine_->blockNotificationFor(30000);
-        auto session = state->session();
-        Bool oldValue = api->get_option(session, option_.c_str());
-        api->set_option(session, option_.c_str(), !oldValue);
-    }
-
-    std::string shortText(InputContext *ic) const override {
-        auto state = engine_->state(ic);
-        auto api = engine_->api();
-        if (!state) {
-            return "";
-        }
-        auto session = state->session();
-        if (api->get_option(session, option_.c_str())) {
-            return stringutils::concat(enabledText_, " → ", disabledText_);
-        }
-        return stringutils::concat(disabledText_, " → ", enabledText_);
-    }
-
-    std::string icon(InputContext *) const override { return ""; }
-
-    std::optional<std::string> snapshotOption(InputContext *ic) override {
-        auto state = engine_->state(ic);
-        auto api = engine_->api();
-        if (!state) {
-            return std::nullopt;
-        }
-        auto session = state->session(false);
-        if (!session) {
-            return std::nullopt;
-        }
-        if (!api->get_option(session, option_.c_str())) {
-            return stringutils::concat("!", option_);
-        }
-        return option_;
-    }
-
-private:
-    RimeEngine *engine_;
-    std::string option_;
-    std::string disabledText_;
-    std::string enabledText_;
-};
-
-class SelectAction : public RimeOptionAction {
-public:
-    SelectAction(RimeEngine *engine, std::string schema,
-                 std::vector<std::string> options,
-                 std::vector<std::string> texts)
-        : engine_(engine), options_(options), texts_(texts) {
-        for (size_t i = 0; i < options.size(); ++i) {
-            actions_.emplace_back();
-            actions_.back().setShortText(texts_[i]);
-            actions_.back().connect<SimpleAction::Activated>(
-                [this, i](InputContext *ic) {
-                    auto state = engine_->state(ic);
-                    auto api = engine_->api();
-                    if (!state) {
-                        return;
-                    }
-                    auto session = state->session();
-                    for (size_t j = 0; j < options_.size(); ++j) {
-                        api->set_option(session, options_[j].c_str(), i == j);
-                    }
-                });
-            engine_->instance()->userInterfaceManager().registerAction(
-                stringutils::concat("fcitx-rime-", schema, "-", options_[i]),
-                &actions_.back());
-            menu_.addAction(&actions_.back());
-        }
-        setMenu(&menu_);
-        engine_->instance()->userInterfaceManager().registerAction(
-            stringutils::concat("fcitx-rime-", schema, "-select-", options[0]),
-            this);
-    }
-
-    std::string shortText(InputContext *ic) const override {
-        auto state = engine_->state(ic);
-        auto api = engine_->api();
-        if (!state) {
-            return "";
-        }
-        auto session = state->session();
-        for (size_t i = 0; i < options_.size(); ++i) {
-            if (api->get_option(session, options_[i].c_str())) {
-                return texts_[i];
-            }
-        }
-        return "";
-    }
-
-    std::string icon(InputContext *) const override { return ""; }
-
-    std::optional<std::string> snapshotOption(InputContext *ic) override {
-        auto state = engine_->state(ic);
-        auto api = engine_->api();
-        if (!state) {
-            return std::nullopt;
-        }
-        auto session = state->session(false);
-        if (!session) {
-            return std::nullopt;
-        }
-        for (size_t i = 0; i < options_.size(); ++i) {
-            if (api->get_option(session, options_[i].c_str())) {
-                return options_[i];
-            }
-        }
-        return std::nullopt;
-    }
-
-private:
-    RimeEngine *engine_;
-    std::vector<std::string> options_;
-    std::vector<std::string> texts_;
-    std::list<SimpleAction> actions_;
-    Menu menu_;
-};
-
 bool RimeEngine::firstRun_ = true;
 
 RimeEngine::RimeEngine(Instance *instance)
     : instance_(instance), api_(EnsureRimeApi()),
       factory_([this](InputContext &ic) { return new RimeState(this, ic); }),
       sessionPool_(this, getSharedStatePolicy()) {
-    if constexpr (isAndroid()) {
+    if constexpr (isAndroid() || isApple()) {
         const auto &sp = fcitx::StandardPath::global();
         std::string defaultYaml = sp.locate(fcitx::StandardPath::Type::Data,
                                             "rime-data/default.yaml");
@@ -314,7 +198,7 @@ RimeEngine::RimeEngine(Instance *instance)
     deployAction_.setShortText(_("Deploy"));
     deployAction_.connect<SimpleAction::Activated>([this](InputContext *ic) {
         deploy();
-        auto state = this->state(ic);
+        auto *state = this->state(ic);
         if (state && ic->hasFocus()) {
             state->updateUI(ic, false);
         }
@@ -327,7 +211,7 @@ RimeEngine::RimeEngine(Instance *instance)
 
     syncAction_.connect<SimpleAction::Activated>([this](InputContext *ic) {
         sync();
-        auto state = this->state(ic);
+        auto *state = this->state(ic);
         if (state && ic->hasFocus()) {
             state->updateUI(ic, false);
         }
@@ -377,7 +261,7 @@ void RimeEngine::rimeStart(bool fullcheck) {
     // make librime only log to stderr
     // https://github.com/rime/librime/commit/6d1b9b65de4e7784a68a17d10a3e5c900e4fd511
     fcitx_rime_traits.log_dir = "";
-    switch (rime().logLevel()) {
+    switch (rime_log().logLevel()) {
     case NoLog:
         fcitx_rime_traits.min_log_level = 4;
         break;
@@ -425,12 +309,14 @@ void RimeEngine::rimeStart(bool fullcheck) {
 
     if (!api_->is_maintenance_mode()) {
         updateAppOptions();
+    } else {
+        needRefreshAppOption_ = true;
     }
 }
 
 void RimeEngine::updateAppOptions() {
     appOptions_.clear();
-    RimeConfig config = {NULL};
+    RimeConfig config = {nullptr};
     if (api_->config_open("fcitx5", &config)) {
         appOptions_ = parseAppOptions(api_, &config);
         api_->config_close(&config);
@@ -444,7 +330,8 @@ void RimeEngine::reloadConfig() {
     updateConfig();
 }
 
-void RimeEngine::setSubConfig(const std::string &path, const RawConfig &) {
+void RimeEngine::setSubConfig(const std::string &path,
+                              const RawConfig & /*unused*/) {
     if (path == "deploy") {
         deploy();
     } else if (path == "sync") {
@@ -516,6 +403,9 @@ void RimeEngine::updateConfig() {
     updateSchemaMenu();
     refreshSessionPoolPolicy();
 
+    deployAction_.setHotkey(config_.deploy.value());
+    syncAction_.setHotkey(config_.synchronize.value());
+
     if (constructed_) {
         refreshStatusArea(0);
     }
@@ -531,7 +421,7 @@ void RimeEngine::refreshStatusArea(InputContext &ic) {
     statusArea.clearGroup(StatusGroup::InputMethod);
     statusArea.addAction(StatusGroup::InputMethod, imAction_.get());
 
-    auto rimeState = state(&ic);
+    auto *rimeState = state(&ic);
     std::string currentSchema;
     if (!rimeState) {
         return;
@@ -554,7 +444,7 @@ void RimeEngine::refreshStatusArea(InputContext &ic) {
 void RimeEngine::refreshStatusArea(RimeSessionId session) {
     instance_->inputContextManager().foreachFocused(
         [this, session](InputContext *ic) {
-            if (auto state = this->state(ic)) {
+            if (auto *state = this->state(ic)) {
                 // After a deployment, param is 0, refresh all
                 if (!session || state->session(false) == session) {
                     refreshStatusArea(*ic);
@@ -570,7 +460,7 @@ void RimeEngine::updateStatusArea(RimeSessionId session) {
             if (instance_->inputMethod(ic) != "rime") {
                 return true;
             }
-            if (auto state = this->state(ic)) {
+            if (auto *state = this->state(ic)) {
                 // After a deployment, param is 0, refresh all
                 if (!session || state->session(false) == session) {
                     // Re-read new option values.
@@ -581,18 +471,33 @@ void RimeEngine::updateStatusArea(RimeSessionId session) {
         });
 }
 
-void RimeEngine::activate(const InputMethodEntry &, InputContextEvent &event) {
-    auto ic = event.inputContext();
+void RimeEngine::activate(const InputMethodEntry & /*entry*/,
+                          InputContextEvent &event) {
+    auto *ic = event.inputContext();
     refreshStatusArea(*ic);
+    if (auto *state = this->state(ic)) {
+        state->activate();
+    }
 }
 
 void RimeEngine::deactivate(const InputMethodEntry &entry,
                             InputContextEvent &event) {
-    if (event.type() == EventType::InputContextSwitchInputMethod &&
-        *config_.commitWhenDeactivate) {
-        auto inputContext = event.inputContext();
-        auto state = inputContext->propertyFor(&factory_);
-        state->commitPreedit(inputContext);
+    if (event.type() == EventType::InputContextSwitchInputMethod) {
+        auto *inputContext = event.inputContext();
+        auto *state = this->state(inputContext);
+        switch (*config_.switchInputMethodBehavior) {
+        case SwitchInputMethodBehavior::Clear:
+            break;
+        case SwitchInputMethodBehavior::CommitRawInput:
+            state->commitInput(inputContext);
+            break;
+        case SwitchInputMethodBehavior::CommitComposingText:
+            state->commitComposing(inputContext);
+            break;
+        case SwitchInputMethodBehavior::CommitCommitPreview:
+            state->commitPreedit(inputContext);
+            break;
+        }
     }
     reset(entry, event);
 }
@@ -602,16 +507,28 @@ void RimeEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
     lastKeyEventTime_ = now(CLOCK_MONOTONIC);
     RIME_DEBUG() << "Rime receive key: " << event.rawKey() << " "
                  << event.isRelease();
-    auto inputContext = event.inputContext();
-    auto state = inputContext->propertyFor(&factory_);
+    auto *inputContext = event.inputContext();
+    if (!event.isRelease()) {
+        if (event.key().checkKeyList(*config_.deploy)) {
+            deploy();
+            return event.filterAndAccept();
+        } else if (event.key().checkKeyList(*config_.synchronize)) {
+            sync();
+            return event.filterAndAccept();
+        }
+    }
+    auto *state = this->state(inputContext);
+    currentKeyEventState_ = state;
     state->keyEvent(event);
+    currentKeyEventState_ = nullptr;
 }
 
-void RimeEngine::reset(const InputMethodEntry &, InputContextEvent &event) {
-    auto inputContext = event.inputContext();
-
-    auto state = inputContext->propertyFor(&factory_);
+void RimeEngine::reset(const InputMethodEntry & /*entry*/,
+                       InputContextEvent &event) {
+    auto *inputContext = event.inputContext();
+    auto *state = this->state(inputContext);
     state->clear();
+    instance_->resetCompose(inputContext);
     inputContext->inputPanel().reset();
     inputContext->updatePreedit();
     inputContext->updateUserInterface(UserInterfaceComponent::InputPanel);
@@ -632,7 +549,10 @@ void RimeEngine::rimeNotificationHandler(void *context, RimeSessionId session,
                                          const char *messageValue) {
     RIME_DEBUG() << "Notification: " << session << " " << messageType << " "
                  << messageValue;
-    RimeEngine *that = static_cast<RimeEngine *>(context);
+    auto *that = static_cast<RimeEngine *>(context);
+    if (that->mainThreadId_ == std::this_thread::get_id()) {
+        that->notifyImmediately(session, messageType, messageValue);
+    }
     that->eventDispatcher_.schedule(
         [that, session, messageType = std::string(messageType),
          messageValue = std::string(messageValue)]() {
@@ -640,12 +560,26 @@ void RimeEngine::rimeNotificationHandler(void *context, RimeSessionId session,
         });
 }
 
+void RimeEngine::notifyImmediately(RimeSessionId session,
+                                   std::string_view messageType,
+                                   std::string_view messageValue) {
+    if (messageType != "option") {
+        return;
+    }
+    if (!currentKeyEventState_ ||
+        currentKeyEventState_->session(false) != session) {
+        return;
+    }
+    currentKeyEventState_->addChangedOption(messageValue);
+}
+
 void RimeEngine::notify(RimeSessionId session, const std::string &messageType,
                         const std::string &messageValue) {
     const char *message = nullptr;
     const char *icon = "";
     const char *tipId = "";
     int timeout = 3000;
+    bool blockMessage = false;
     if (messageType == "deploy") {
         tipId = "fcitx-rime-deploy";
         icon = "fcitx_rime_deploy";
@@ -655,52 +589,38 @@ void RimeEngine::notify(RimeSessionId session, const std::string &messageType,
         } else if (messageValue == "success") {
             message = _("Rime is ready.");
             if (!api_->is_maintenance_mode()) {
-                api_->deploy_config_file("fcitx5.yaml", "config_version");
-                updateAppOptions();
+                if (needRefreshAppOption_) {
+                    api_->deploy_config_file("fcitx5.yaml", "config_version");
+                    updateAppOptions();
+                    needRefreshAppOption_ = false;
+                }
             }
             updateSchemaMenu();
             refreshStatusArea(0);
+            blockMessage = true;
         } else if (messageValue == "failure") {
+            needRefreshAppOption_ = false;
             message = _("Rime has encountered an error. "
                         "See log for details.");
+            blockMessage = true;
         }
     } else if (messageType == "option") {
-        // Only show option notification triggered by key event.
-        if (!isAndroid() && lastKeyEventTime_ + 30000 > now(CLOCK_MONOTONIC)) {
-            icon = "fcitx-rime";
-            if (messageValue == "!full_shape") {
-                tipId = "fcitx-rime-full-shape";
-                message = _("Half Shape is enabled.");
-            } else if (messageValue == "full_shape") {
-                tipId = "fcitx-rime-full-shape";
-                message = _("Full Shape is enabled.");
-            } else if (messageValue == "!ascii_punct") {
-                tipId = "fcitx-rime-ascii-punct";
-                message = _("Punctuation conversion is enabled.");
-            } else if (messageValue == "ascii_punct") {
-                tipId = "fcitx-rime-ascii-punct";
-                message = _("Punctuation conversion is disabled.");
-            } else if (messageValue == "!simplification") {
-                tipId = "fcitx-rime-simplification";
-                message = _("Traditional Chinese is enabled.");
-            } else if (messageValue == "simplification") {
-                tipId = "fcitx-rime-simplification";
-                message = _("Simplified Chinese is enabled.");
-            }
-        }
-        timeout = 1000;
         updateStatusArea(session);
     } else if (messageType == "schema") {
         // Schema is changed either via status area or shortcut
         refreshStatusArea(session);
     }
 
-    auto notifications = this->notifications();
+    auto *notifications = this->notifications();
     if (message && notifications &&
         now(CLOCK_MONOTONIC) > blockNotificationBefore_) {
         notifications->call<INotifications::showTip>(
             tipId, _("Rime"), icon, _("Rime"), message, timeout);
     }
+    // Block message after error / success.
+    if (blockMessage) {
+        blockNotificationFor(30000);
+    }
 }
 
 RimeState *RimeEngine::state(InputContext *ic) {
@@ -710,28 +630,29 @@ RimeState *RimeEngine::state(InputContext *ic) {
     return ic->propertyFor(&factory_);
 }
 
-std::string RimeEngine::subMode(const InputMethodEntry &, InputContext &ic) {
-    if (auto rimeState = state(&ic)) {
+std::string RimeEngine::subMode(const InputMethodEntry & /*entry*/,
+                                InputContext &ic) {
+    if (auto *rimeState = state(&ic)) {
         return rimeState->subMode();
     }
     return "";
 }
 
-std::string RimeEngine::subModeLabelImpl(const InputMethodEntry &,
+std::string RimeEngine::subModeLabelImpl(const InputMethodEntry & /*unused*/,
                                          InputContext &ic) {
-    if (auto rimeState = state(&ic)) {
+    if (auto *rimeState = state(&ic)) {
         return rimeState->subModeLabel();
     }
     return "";
 }
 
-std::string RimeEngine::subModeIconImpl(const InputMethodEntry &,
+std::string RimeEngine::subModeIconImpl(const InputMethodEntry & /*unused*/,
                                         InputContext &ic) {
     std::string result = "fcitx-rime";
     if (!factory_.registered()) {
         return result;
     }
-    auto state = this->state(&ic);
+    auto *state = this->state(&ic);
     if (state) {
         state->getStatus([&result](const RimeStatus &status) {
             if (status.is_disabled) {
@@ -746,9 +667,9 @@ std::string RimeEngine::subModeIconImpl(const InputMethodEntry &,
     return result;
 }
 
-void RimeEngine::releaseAllSession(const bool snapshot) {
+void RimeEngine::releaseAllSession(bool snapshot) {
     instance_->inputContextManager().foreach([&](InputContext *ic) {
-        if (auto state = this->state(ic)) {
+        if (auto *state = this->state(ic)) {
             if (snapshot) {
                 state->snapshot();
             }
@@ -784,7 +705,7 @@ void RimeEngine::updateActionsForSchema(const std::string &schema) {
             continue;
         }
         auto namePath = switchPath + "/name";
-        auto name = api_->config_get_cstring(&config, namePath.c_str());
+        const auto *name = api_->config_get_cstring(&config, namePath.c_str());
         if (name) {
             if (labels.size() != 2) {
                 continue;
@@ -822,8 +743,8 @@ void RimeEngine::updateSchemaMenu() {
         schemActions_.back().setShortText(_("Latin Mode"));
         schemActions_.back().connect<SimpleAction::Activated>(
             [this](InputContext *ic) {
-                auto state = ic->propertyFor(&factory_);
-                state->setLatinMode(true);
+                auto *state = this->state(ic);
+                state->toggleLatinMode();
                 imAction_->update(ic);
             });
         instance_->userInterfaceManager().registerAction(&schemActions_.back());
@@ -835,7 +756,7 @@ void RimeEngine::updateSchemaMenu() {
             schemaAction.setShortText(list.list[i].name);
             schemaAction.connect<SimpleAction::Activated>(
                 [this, schemaId](InputContext *ic) {
-                    auto state = ic->propertyFor(&factory_);
+                    auto *state = this->state(ic);
                     blockNotificationFor(30000);
                     state->selectSchema(schemaId);
                     imAction_->update(ic);
@@ -871,6 +792,4 @@ PropertyPropagatePolicy RimeEngine::getSharedStatePolicy() {
     }
 }
 
-} // namespace fcitx
-
-FCITX_ADDON_FACTORY(fcitx::RimeEngineFactory)
+} // namespace fcitx::rime
